#include <stdint.h>
#include <stdlib.h>

#include "base64.h"

static char encoding_table[] = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                                'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
                                'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
                                'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
                                'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
                                'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
                                'w', 'x', 'y', 'z', '0', '1', '2', '3',
                                '4', '5', '6', '7', '8', '9', '+', '/'};
static char *decoding_table = NULL;
static int mod_table[] = {0, 2, 1};


char *base64_encode(const char *data, int input_length, int *output_length) {
  char *encoded_data;
  uint32_t octet_a, octet_b, octet_c;
  uint32_t triple; 
  int i, j;

  *output_length = ((input_length - 1) / 3) * 4 + 4;
  
  encoded_data = (char *)malloc(*output_length);
  if (!encoded_data) return NULL;

  for (i = 0, j = 0; i < input_length;) {

    octet_a = i < input_length ? data[i++] : 0;
    octet_b = i < input_length ? data[i++] : 0;
    octet_c = i < input_length ? data[i++] : 0;
    
    triple = (octet_a << 0x10) + (octet_b << 0x08) + octet_c;
    
    encoded_data[j++] = encoding_table[(triple >> 3 * 6) & 0x3F];
    encoded_data[j++] = encoding_table[(triple >> 2 * 6) & 0x3F];
    encoded_data[j++] = encoding_table[(triple >> 1 * 6) & 0x3F];
    encoded_data[j++] = encoding_table[(triple >> 0 * 6) & 0x3F];
  }

  for (i = 0; i < mod_table[input_length % 3]; i++)
    encoded_data[*output_length - 1 - i] = '=';

  encoded_data[*output_length - i] = '\0'; // NULL-terminate the string

  return encoded_data;
}


char *base64_decode(const char *data, int input_length, int *output_length) {
  char *decoded_data;
  uint32_t sextet_a, sextet_b, sextet_c, sextet_d; 
  uint32_t triple;
  int in_cnt, out_cnt, k;
  
  if (decoding_table == NULL) base64_build_decoding_table();
 
  // there might be leading whitespace...
  //  k = 0;
  //  while ((data[k] == 13) || (data[k] == 10)) k++;
  
  //  if (input_length % 4 != 0) { 
  //    *output_length = 0;
  //    return NULL;
  //  }
  
  *output_length = input_length / 4 * 3;
  //  if (data[input_length - 1] == '=') (*output_length)--;
  //  if (data[input_length - 2] == '=') (*output_length)--;

  decoded_data = (char *)malloc(*output_length);
  if (!decoded_data) {
    *output_length = 0;
    return NULL;
  }
  
  out_cnt = 0;
  in_cnt  = 0;
  while (input_length >= in_cnt) {

    while (((data[in_cnt] == 13) || (data[in_cnt] == 10)) && (input_length >= in_cnt)) in_cnt++;
    if (input_length < in_cnt) break;
    sextet_a = data[in_cnt] == '=' ? 0 & in_cnt++ : decoding_table[data[in_cnt++]];

    while (((data[in_cnt] == 13) || (data[in_cnt] == 10)) && (input_length >= in_cnt)) in_cnt++;
    if (input_length < in_cnt) break;
    sextet_b = data[in_cnt] == '=' ? 0 & in_cnt++ : decoding_table[data[in_cnt++]];

    while (((data[in_cnt] == 13) || (data[in_cnt] == 10)) && (input_length >= in_cnt)) in_cnt++;
    if (input_length < in_cnt) break;
    sextet_c = data[in_cnt] == '=' ? 0 & in_cnt++ : decoding_table[data[in_cnt++]];

    while (((data[in_cnt] == 13) || (data[in_cnt] == 10)) && (input_length >= in_cnt)) in_cnt++;
    if (input_length < in_cnt) break;
    sextet_d = data[in_cnt] == '=' ? 0 & in_cnt++ : decoding_table[data[in_cnt++]];

    triple = (sextet_a << 3 * 6) + (sextet_b << 2 * 6) + (sextet_c << 1 * 6) + (sextet_d << 0 * 6);
    
    if (out_cnt < *output_length) decoded_data[out_cnt++] = (triple >> 2 * 8) & 0xFF; 
    if (out_cnt < *output_length) decoded_data[out_cnt++] = (triple >> 1 * 8) & 0xFF; 
    if (out_cnt < *output_length) decoded_data[out_cnt++] = (triple >> 0 * 8) & 0xFF; 
  }

  decoded_data[out_cnt++] = '\0'; // NULL-terminate the string
  *output_length = out_cnt;

  return decoded_data;
}


void base64_build_decoding_table() {
  int i;
  decoding_table = malloc(256);
  
  for (i = 0; i < 0x40; i++)
    decoding_table[encoding_table[i]] = i;
}


void base64_cleanup() {
  free(decoding_table);
}
