#include <stdio.h>
#include <math.h>

#include "helpers.h"
#include "data_handling.h" 
#include "base64.h"

 struct t_EGG_NOM_2_data {
  long double time;                    // Time_Information, GPS_Time
  double Vxx, Vyy, Vzz, Vxy, Vxz, Vyz; // Gravity_Gradients
  double Sxx, Syy, Szz, Sxy, Sxz, Syz; // Sigmas
  short int Fxx, Fyy, Fzz, Fxy, Fxz, Fyz; // Flags
  double TTDxx, TTDyy, TTDzz, TTDxy, TTDxz, TTDyz; // Corrections, Temporal, Tidal, Direct_Tides
  double TTSxx, TTSyy, TTSzz, TTSxy, TTSxz, TTSyz; // Corrections, Temporal, Tidal, Solid_Earth
  double TTOxx, TTOyy, TTOzz, TTOxy, TTOxz, TTOyz; // Corrections, Temporal, Tidal, Ocean_Tides
  double TTPxx, TTPyy, TTPzz, TTPxy, TTPxz, TTPyz; // Corrections, Temporal, Tidal, Polen_Tides
  double TNxx, TNyy, TNzz, TNxy, TNxz, TNyz;       // Corrections, Temporal, Non-Tidal
  double Cxx, Cyy, Czz, Cxy, Cxz, Cyz;             // Corrections, Calibration
  double Q1, Q2, Q3, Q4;               // Quaternions
} EGG_NOM_2_data;

struct t_SST_PCV_2_data {
  int row, column;                // row, column of covariance matrix
  double Cov;                     // value 
} SST_PCV_2_data;

struct t_SST_PKI_2_data {
  double YY, MM, DoM, hh, mm, ss; // time
  double Position_Factor;
  double X, Y, Z, Clk;            // position, clock
} SST_PKI_2_data;

struct t_SST_PRD_2_data {
  double YY, MM, DoM, hh, mm, ss; // time
  double Position_Factor;
  double X, Y, Z, Clk;            // position, clock
  double Velocity_Factor;
  double Xd, Yd, Zd, Clkd;        // changes of position, clock
} SST_PRD_2_data;

struct t_SST_PRM_2_data {
  double reftime;                 // time
  double Q1, Q2, Q3, Q4;          // quaternions
} SST_PRM_2_data;

struct t_HDR_SST_PRM_2_data {
  double refepoch;                // reference epoch
} HDR_SST_PRM_2_data;

struct t_SST_PRP_2_data {
  int encoding;                   // encoding
} SST_PRP_2_data;



/**********************************************************/


void initialize_EGG_NOM_2_data() {
  EGG_NOM_2_data.time = NAN;
  EGG_NOM_2_data.Vxx = NAN; EGG_NOM_2_data.Vyy = NAN; EGG_NOM_2_data.Vzz = NAN;
  EGG_NOM_2_data.Vxy = NAN; EGG_NOM_2_data.Vxz = NAN; EGG_NOM_2_data.Vyz = NAN;
  EGG_NOM_2_data.Sxx = NAN; EGG_NOM_2_data.Syy = NAN; EGG_NOM_2_data.Szz = NAN; 
  EGG_NOM_2_data.Sxy = NAN; EGG_NOM_2_data.Sxz = NAN; EGG_NOM_2_data.Syz = NAN;
  EGG_NOM_2_data.Fxx = -1; EGG_NOM_2_data.Fyy = -1; EGG_NOM_2_data.Fzz = -1;
  EGG_NOM_2_data.Fxy = -1; EGG_NOM_2_data.Fxz = -1; EGG_NOM_2_data.Fyz = -1;
  EGG_NOM_2_data.TTDxx = NAN; EGG_NOM_2_data.TTDyy = NAN; EGG_NOM_2_data.TTDzz = NAN; 
  EGG_NOM_2_data.TTDxy = NAN; EGG_NOM_2_data.TTDxz = NAN; EGG_NOM_2_data.TTDyz = NAN; 
  EGG_NOM_2_data.TTSxx = NAN; EGG_NOM_2_data.TTSyy = NAN; EGG_NOM_2_data.TTSzz = NAN;
  EGG_NOM_2_data.TTSxy = NAN; EGG_NOM_2_data.TTSxz = NAN; EGG_NOM_2_data.TTSyz = NAN; 
  EGG_NOM_2_data.TTOxx = NAN; EGG_NOM_2_data.TTOyy = NAN; EGG_NOM_2_data.TTOzz = NAN;
  EGG_NOM_2_data.TTOxy = NAN; EGG_NOM_2_data.TTOxz = NAN; EGG_NOM_2_data.TTOyz = NAN; 
  EGG_NOM_2_data.TTPxx = NAN; EGG_NOM_2_data.TTPyy = NAN; EGG_NOM_2_data.TTPzz = NAN;
  EGG_NOM_2_data.TTPxy = NAN; EGG_NOM_2_data.TTPxz = NAN; EGG_NOM_2_data.TTPyz = NAN; 
  EGG_NOM_2_data.TNxx = NAN; EGG_NOM_2_data.TNyy = NAN; EGG_NOM_2_data.TNzz = NAN;
  EGG_NOM_2_data.TNxy = NAN; EGG_NOM_2_data.TNxz = NAN; EGG_NOM_2_data.TNyz = NAN;       
  EGG_NOM_2_data.Cxx = NAN; EGG_NOM_2_data.Cyy = NAN; EGG_NOM_2_data.Czz = NAN;
  EGG_NOM_2_data.Cxy = NAN; EGG_NOM_2_data.Cxz = NAN; EGG_NOM_2_data.Cyz = NAN;
  EGG_NOM_2_data.Q1 = NAN; EGG_NOM_2_data.Q2 = NAN; EGG_NOM_2_data.Q3 = NAN; EGG_NOM_2_data.Q4 = NAN;
} 

void initialize_SST_PCV_2_data() {
  SST_PCV_2_data.row = -999999; SST_PCV_2_data.column = -999999; SST_PCV_2_data.Cov = NAN; 
} 

void initialize_SST_PKI_2_data() {
  SST_PKI_2_data.YY = NAN; SST_PKI_2_data.MM = NAN; SST_PKI_2_data.DoM = NAN; 
  SST_PKI_2_data.hh = NAN; SST_PKI_2_data.mm = NAN; SST_PKI_2_data.ss  = NAN;
  SST_PKI_2_data.Position_Factor = 1.0;
  SST_PKI_2_data.X  = NAN; SST_PKI_2_data.Y  = NAN; SST_PKI_2_data.Z   = NAN; 
  SST_PKI_2_data.Clk = NAN; 
}

void initialize_SST_PRD_2_data() {
  SST_PRD_2_data.YY = NAN; SST_PRD_2_data.MM = NAN; SST_PRD_2_data.DoM = NAN; 
  SST_PRD_2_data.hh = NAN; SST_PRD_2_data.mm = NAN; SST_PRD_2_data.ss  = NAN;
  SST_PRD_2_data.Position_Factor = 1.0;
  SST_PRD_2_data.X  = NAN; SST_PRD_2_data.Y  = NAN; SST_PRD_2_data.Z   = NAN; SST_PRD_2_data.Clk  = NAN; 
  SST_PRD_2_data.Velocity_Factor = 1.0;
  SST_PRD_2_data.Xd = NAN; SST_PRD_2_data.Yd = NAN; SST_PRD_2_data.Zd  = NAN; SST_PRD_2_data.Clkd = NAN; 
} 

void initialize_SST_PRM_2_data() {
  SST_PRM_2_data.reftime = NAN;
  SST_PRM_2_data.Q1 = NAN; SST_PRM_2_data.Q2 = NAN; SST_PRM_2_data.Q3 = NAN; SST_PRM_2_data.Q4 = NAN;
} 

void HDR_initialize_SST_PRM_2_data() {
  HDR_SST_PRM_2_data.refepoch = 0.0;
}


/************************************************************/


void PutToStruct_EGG_NOM_2(struct t_tag_flags tag, const xmlChar *value) {
  int err = -1;

  switch (tag.level_0) {
  case TAG_0_Data_Block:

    switch (tag.level_1) {
    case TAG_1_EGG_NOM_2:

      switch (tag.level_2) {
      case TAG_2_List_of_GG_time_Records:

	switch (tag.level_3) {
	case TAG_3_GG_time_Record:

	  switch (tag.level_4) {
	  case TAG_4_Time_Information: 
	    switch (tag.level_5) {
	    case TAG_5_GPS_Time: EGG_NOM_2_data.time = atof(value); break;
	    default: err = 5;
	    } 
	    break; // TAG_4_Time_Information

	  case TAG_4_Gravity_Gradients:
	    switch (tag.level_5) {
	    case TAG_5_XX: EGG_NOM_2_data.Vxx = atof(value); break;
	    case TAG_5_YY: EGG_NOM_2_data.Vyy = atof(value); break;
	    case TAG_5_ZZ: EGG_NOM_2_data.Vzz = atof(value); break;
	    case TAG_5_XY: EGG_NOM_2_data.Vxy = atof(value); break;
	    case TAG_5_XZ: EGG_NOM_2_data.Vxz = atof(value); break;
	    case TAG_5_YZ: EGG_NOM_2_data.Vyz = atof(value); break;
	    default: err = 5;
	    }
	    break; // TAG_4_Gravity_Gradients

	  case TAG_4_Sigmas:
	    switch (tag.level_5) {
	    case TAG_5_XX: EGG_NOM_2_data.Sxx = atof(value); break;
	    case TAG_5_YY: EGG_NOM_2_data.Syy = atof(value); break;
	    case TAG_5_ZZ: EGG_NOM_2_data.Szz = atof(value); break;
	    case TAG_5_XY: EGG_NOM_2_data.Sxy = atof(value); break;
	    case TAG_5_XZ: EGG_NOM_2_data.Sxz = atof(value); break;
	    case TAG_5_YZ: EGG_NOM_2_data.Syz = atof(value); break;
	    default: err = 5;
	    }
	    break; // TAG_4_Sigmas

	  case TAG_4_Flags:
	    switch (tag.level_5) {
	    case TAG_5_XX: EGG_NOM_2_data.Fxx = atoi(value); break;
	    case TAG_5_YY: EGG_NOM_2_data.Fyy = atoi(value); break;
	    case TAG_5_ZZ: EGG_NOM_2_data.Fzz = atoi(value); break;
	    case TAG_5_XY: EGG_NOM_2_data.Fxy = atoi(value); break;
	    case TAG_5_XZ: EGG_NOM_2_data.Fxz = atoi(value); break;
	    case TAG_5_YZ: EGG_NOM_2_data.Fyz = atoi(value); break;
	    default: err = 5;
	    }
	    break; // TAG_4_Flags

	  case TAG_4_Corrections:
	    switch (tag.level_5) {
	    case TAG_5_Temporal:
	      switch (tag.level_6) {
	      case TAG_6_Tidal:
		switch (tag.level_7) {
		case TAG_7_Direct_Tides:
		  switch (tag.level_8) {
		  case TAG_8_XX: EGG_NOM_2_data.TTDxx = atof(value); break;
		  case TAG_8_YY: EGG_NOM_2_data.TTDyy = atof(value); break;
		  case TAG_8_ZZ: EGG_NOM_2_data.TTDzz = atof(value); break;
		  case TAG_8_XY: EGG_NOM_2_data.TTDxy = atof(value); break;
		  case TAG_8_XZ: EGG_NOM_2_data.TTDxz = atof(value); break;
		  case TAG_8_YZ: EGG_NOM_2_data.TTDyz = atof(value); break;
		  default: err = 8;
		  }
		  break; // TAG_7_Direct_Tides

		case TAG_7_Solid_Earth:
		  switch (tag.level_8) {
		  case TAG_8_XX: EGG_NOM_2_data.TTSxx = atof(value); break;
		  case TAG_8_YY: EGG_NOM_2_data.TTSyy = atof(value); break;
		  case TAG_8_ZZ: EGG_NOM_2_data.TTSzz = atof(value); break;
		  case TAG_8_XY: EGG_NOM_2_data.TTSxy = atof(value); break;
		  case TAG_8_XZ: EGG_NOM_2_data.TTSxz = atof(value); break;
		  case TAG_8_YZ: EGG_NOM_2_data.TTSyz = atof(value); break;
		  default: err = 8;
		  }
		  break; // TAG_7_Solid_Earth

		case TAG_7_Ocean_Tides:
		  switch (tag.level_8) {
		  case TAG_8_XX: EGG_NOM_2_data.TTOxx = atof(value); break;
		  case TAG_8_YY: EGG_NOM_2_data.TTOyy = atof(value); break;
		  case TAG_8_ZZ: EGG_NOM_2_data.TTOzz = atof(value); break;
		  case TAG_8_XY: EGG_NOM_2_data.TTOxy = atof(value); break;
		  case TAG_8_XZ: EGG_NOM_2_data.TTOxz = atof(value); break;
		  case TAG_8_YZ: EGG_NOM_2_data.TTOyz = atof(value); break;
		  default: err = 8;
		  }
		  break; // TAG_7_Ocean_Tides

		case TAG_7_Pole_Tides:
		  switch (tag.level_8) {
		  case TAG_8_XX: EGG_NOM_2_data.TTPxx = atof(value); break;
		  case TAG_8_YY: EGG_NOM_2_data.TTPyy = atof(value); break;
		  case TAG_8_ZZ: EGG_NOM_2_data.TTPzz = atof(value); break;
		  case TAG_8_XY: EGG_NOM_2_data.TTPxy = atof(value); break;
		  case TAG_8_XZ: EGG_NOM_2_data.TTPxz = atof(value); break;
		  case TAG_8_YZ: EGG_NOM_2_data.TTPyz = atof(value); break;
		  default: err = 8;
		  } 
		  break; // TAG_7_Pole_Tides
		default: err = 7;
		} 
		break; // TAG_6_Tidal

	      case TAG_6_Non_Tidal:
		switch (tag.level_7) {
		case TAG_7_XX: EGG_NOM_2_data.TNxx = atof(value); break;
		case TAG_7_YY: EGG_NOM_2_data.TNyy = atof(value); break;
		case TAG_7_ZZ: EGG_NOM_2_data.TNzz = atof(value); break;
		case TAG_7_XY: EGG_NOM_2_data.TNxy = atof(value); break;
		case TAG_7_XZ: EGG_NOM_2_data.TNxz = atof(value); break;
		case TAG_7_YZ: EGG_NOM_2_data.TNyz = atof(value); break;
		default: err = 7;
		}
		break; // TAG_6_Non_Tidal

	      default: err = 6;
	      } 
	      break; // TAG_5_Temporal

	    case TAG_5_Calibration:
	      switch (tag.level_6) {
	      case TAG_6_XX: EGG_NOM_2_data.Cxx = atof(value); break;
	      case TAG_6_YY: EGG_NOM_2_data.Cyy = atof(value); break;
	      case TAG_6_ZZ: EGG_NOM_2_data.Czz = atof(value); break;
	      case TAG_6_XY: EGG_NOM_2_data.Cxy = atof(value); break;
	      case TAG_6_XZ: EGG_NOM_2_data.Cxz = atof(value); break;
	      case TAG_6_YZ: EGG_NOM_2_data.Cyz = atof(value); break;
	      default: err = 6;
	      }
	      break; // TAG_5_Calibration
	    default: err = 5;
	    }
	    break; // TAG_4_Corrections

	  case TAG_4_Quaternions:
	    switch (tag.level_5) {
	    case TAG_5_Q1: EGG_NOM_2_data.Q1 = atof(value); break;
	    case TAG_5_Q2: EGG_NOM_2_data.Q2 = atof(value); break;
	    case TAG_5_Q3: EGG_NOM_2_data.Q3 = atof(value); break;
	    case TAG_5_Q4: EGG_NOM_2_data.Q4 = atof(value); break;
	    default: err = 5;
	    } 
	    break; // TAG_4_Quaternions
	  default: err = 4;
	  }
	  break; // TAG_3_GG_time_Record
	default: err = 3;
	} 
	break; // TAG_2_List_of_GG_time_Records
      default: err = 2;
      } 
      break; // TAG_1_EGG_NOM_2
    default: err = 1;
    } 
    break; // TAG_0_Data_Block
  default: err = 0;
  }
    
  if (err != -1) {
    printf("MESSAGE: tag not found --> level %d, tags %d %d %d %d %d %d %d %d\n", 
	   err, tag.level_0, tag.level_1, tag.level_2, tag.level_3, tag.level_4, tag.level_5, tag.level_6, tag.level_7, tag.level_8);
  }
}

void PutToStruct_SST_PCV_2(struct t_tag_flags tag, int extratag, const xmlChar *row, const xmlChar *column, const xmlChar *value) {
  if ((tag.level_0 == TAG_0_Data_Block) &&
      (tag.level_1 == TAG_1_SST_PSO_2) &&
      (tag.level_2 == TAG_2_SST_PCV_2) &&
      (tag.level_3 == TAG_3_List_of_Covariance_Records) &&
      (tag.level_4 == TAG_4_Covariance_Record) &&
      (tag.level_5 == TAG_5_NONE)) {
    switch (extratag) {
    case TAG_extra_ROW_COLUMN:
      SST_PCV_2_data.row    = atoi(row); 
      SST_PCV_2_data.column = atoi(column); 
      break;
    case TAG_extra_VALUE:
      SST_PCV_2_data.Cov    = atof(value); 
      break;
    } // switch (extratag)
  }
}

void PutToStruct_SST_PKI_2(struct t_tag_flags tag, const xmlChar *value) {
  int err = -1;

  switch (tag.level_0) {
  case TAG_0_Data_Block:
    
    switch (tag.level_1) {
    case TAG_1_SST_PSO_2:

      switch (tag.level_2) {
      case TAG_2_SST_PKI_2:
	
	switch (tag.level_3) {
	case TAG_3_List_of_SP3c_Records:

	  switch (tag.level_4) {
	  case TAG_4_SP3c_Record:
	    
	    switch (tag.level_5) {
	    case TAG_5_Time_Information:

	      switch (tag.level_6) {
	      case TAG_6_GPS_Time:

		switch (tag.level_7) {
		case TAG_7_Start:

		  switch (tag.level_8) {
		  case TAG_8_Gregorian:

		    switch (tag.level_9) {
		    case TAG_9_Year:         SST_PKI_2_data.YY  = atof(value); break;
		    case TAG_9_Month:        SST_PKI_2_data.MM  = atof(value); break;
		    case TAG_9_Day_of_Month: SST_PKI_2_data.DoM = atof(value); break;
		    case TAG_9_Hour:         SST_PKI_2_data.hh  = atof(value); break;
		    case TAG_9_Minute:       SST_PKI_2_data.mm  = atof(value); break;
		    case TAG_9_Second:       SST_PKI_2_data.ss  = atof(value); break;
		    default: err = 9;
		    }
		    break; // TAG_8_Gregorian
		  default: err = 8;
		  }
		  break; // TAG_7_Start
		default: err = 7;
		}
		break; // TAG_6_GPS_Time
	      default: err = 6;
	      }
	      break; // TAG_5_Time_Information
	      
	    case TAG_5_List_of_Satellite_IDs:
	      switch (tag.level_6) {
	      case TAG_6_L15:
		switch (tag.level_7) {
		case TAG_7_Position:
		  switch (tag.level_8) {
		  case TAG_8_X: SST_PKI_2_data.X = atof(value); break;
		  case TAG_8_Y: SST_PKI_2_data.Y = atof(value); break;
		  case TAG_8_Z: SST_PKI_2_data.Z = atof(value); break;
		  default: err = 8;
		  }
		  break;

		case TAG_7_Clock: SST_PKI_2_data.Clk = atof(value); break;
		default: err = 7;
		}
		break; // TAG_6_L15
	      default: err = 6;
	      }
	      break; // TAG_5_List_of_Satellite_IDs
	    default: err = 5;
	    }
	    break; // TAG_4_SP3c_Record
	  default: err = 4;
	  }
	  break; // TAG_3_List_of_SP3c_Records
	default: err = 3;
	}
	break; // TAG_2_SST_PKI_2
      default: err = 2;
      }
      break; // TAG_1_SST_PSO_2
    default: err = 1;
    }
    break; // TAG_0_Data_Block
  default: err = 0;
  }
}

void PutToStruct_SST_PKI_2_Position_Factor(double f) {
  SST_PKI_2_data.Position_Factor = f;
}

void PutToStruct_SST_PRD_2(struct t_tag_flags tag, const xmlChar *value) {
  int err = -1;

  switch (tag.level_0) {
  case TAG_0_Data_Block:
    
    switch (tag.level_1) {
    case TAG_1_SST_PSO_2:

      switch (tag.level_2) {
      case TAG_2_SST_PRD_2:
	
	switch (tag.level_3) {
	case TAG_3_List_of_SP3c_Records:

	  switch (tag.level_4) {
	  case TAG_4_SP3c_Record:
	    
	    switch (tag.level_5) {
	    case TAG_5_Time_Information:

	      switch (tag.level_6) {
	      case TAG_6_GPS_Time:

		switch (tag.level_7) {
		case TAG_7_Start:

		  switch (tag.level_8) {
		  case TAG_8_Gregorian:

		    switch (tag.level_9) {
		    case TAG_9_Year:         SST_PRD_2_data.YY  = atof(value); break;
		    case TAG_9_Month:        SST_PRD_2_data.MM  = atof(value); break;
		    case TAG_9_Day_of_Month: SST_PRD_2_data.DoM = atof(value); break;
		    case TAG_9_Hour:         SST_PRD_2_data.hh  = atof(value); break;
		    case TAG_9_Minute:       SST_PRD_2_data.mm  = atof(value); break;
		    case TAG_9_Second:       SST_PRD_2_data.ss  = atof(value); break;
		    default: err = 9;
		    }
		    break; // TAG_8_Gregorian
		  default: err = 8;
		  }
		  break; // TAG_7_Start
		default: err = 7;
		}
		break; // TAG_6_GPS_Time
	      default: err = 6;
	      }
	      break; // TAG_5_Time_Information
	      
	    case TAG_5_List_of_Satellite_IDs:
	      switch (tag.level_6) {
	      case TAG_6_L15:
		switch (tag.level_7) {
		case TAG_7_Position:
		  switch (tag.level_8) {
		  case TAG_8_X: SST_PRD_2_data.X = atof(value); break;
		  case TAG_8_Y: SST_PRD_2_data.Y = atof(value); break;
		  case TAG_8_Z: SST_PRD_2_data.Z = atof(value); break;
		  default: err = 8;
		  }
		  break;
		  
		case TAG_7_Velocity:
		  switch (tag.level_8) {
		  case TAG_8_X: SST_PRD_2_data.Xd = atof(value); break;
		  case TAG_8_Y: SST_PRD_2_data.Yd = atof(value); break;
		  case TAG_8_Z: SST_PRD_2_data.Zd = atof(value); break;
		  default: err = 8;
		  }
		  break;

		case TAG_7_Clock: SST_PRD_2_data.Clk = atof(value); break;
		case TAG_7_Clock_Change_Rate: SST_PRD_2_data.Clkd = atof(value); break;

		default: err = 7;
		}
		break; // TAG_6_L15
	      default: err = 6;
	      }
	      break; // TAG_5_List_of_Satellite_IDs
	    default: err = 5;
	    }
	    break; // TAG_4_SP3c_Record
	  default: err = 4;
	  }
	  break; // TAG_3_List_of_SP3c_Records
	default: err = 3;
	}
	break; // TAG_2_SST_PRD_2
      default: err = 2;
      }
      break; // TAG_1_SST_PSO_2
    default: err = 1;
    }
    break; // TAG_0_Data_Block
  default: err = 0;
  }
}

void PutToStruct_SST_PRD_2_Position_Factor(double f) {
  SST_PRD_2_data.Position_Factor = f;
}

void PutToStruct_SST_PRD_2_Velocity_Factor(double f) {
  SST_PRD_2_data.Velocity_Factor = f;
}


void PutToStruct_SST_PRM_2(struct t_tag_flags tag, const xmlChar *value) {
  int err = -1;

  if ((tag.level_0 == TAG_0_Data_Block) &&
      (tag.level_1 == TAG_1_SST_PSO_2) &&
      (tag.level_2 == TAG_2_SST_PRM_2) &&
      (tag.level_3 == TAG_3_List_of_Rotation_Records) &&
      (tag.level_4 == TAG_4_Rotation_Record)) {
    switch (tag.level_5) {
    case TAG_5_Time_Information:
      switch (tag.level_6) {
      case TAG_6_Time_Since_Reference_Epoch: 
	SST_PRM_2_data.reftime = HDR_SST_PRM_2_data.refepoch + atof(value);
	break;
      default: err = 6;
      }
      
    case TAG_5_Quaternions:
      switch (tag.level_6) {
      case TAG_6_Q1: SST_PRM_2_data.Q1 = atof(value); break;
      case TAG_6_Q2: SST_PRM_2_data.Q2 = atof(value); break;
      case TAG_6_Q3: SST_PRM_2_data.Q3 = atof(value); break;
      case TAG_6_Q4: SST_PRM_2_data.Q4 = atof(value); break;
      default: err = 6;
      }
      
    default: err = 5;
    }
  } else { 
    err = 0;
  }
} 

void PutToStruct_SST_PRP_2(struct t_tag_flags tag, const xmlChar *value) {
  if ((tag.level_0 == TAG_0_Data_Block) &&
      (tag.level_1 == TAG_1_SST_PSO_2) &&
      (tag.level_2 == TAG_2_SST_PRP_2) &&
      (tag.level_3 == TAG_3_List_of_PDF_Files) &&
      (tag.level_4 == TAG_4_PDF_File) &&
      (tag.level_5 == TAG_5_NONE)) {
    if (strcmp(value, "Base64") == 0)
      SST_PRP_2_data.encoding = ENCODING_BASE64;
    else 
      SST_PRP_2_data.encoding = ENCODING_NONE;
  }
} 

void HDR_PutToStruct_SST_PRM_2(unsigned int tag, const xmlChar *value) {
  double Y, M, D, h, m, s;
  
  switch (tag) {
  case HDR_TAG_Earth_Explorer_Header | HDR_TAG_Variable_Header | HDR_TAG_SPH | HDR_TAG_SST_PSO_2 | 
    HDR_TAG_SST_PRM_2 | HDR_TAG_Epoch_Information | HDR_TAG_Reference:
    sscanf(value, "%lf-%lf-%lf %lf:%lf:%lf GPS", &Y, &M, &D, &h, &m, &s);
    HDR_SST_PRM_2_data.refepoch = (greg2jul(Y, M, D, h, m, s) - greg2jul(1980.0, 1.0, 6.0, 0.0, 0.0, 0.0)) * 24 * 3600;
    
    printf("found epoch reference: %4.0f %2.0f %2.0f %2.0f %2.0f %2.0f (in GPS-time: %f)\n", Y, M, D, h, m, s, HDR_SST_PRM_2_data.refepoch);
    break;
  }
} 

void PutToStruct_SST_PRP_2_Encoding(int encoding) {
  SST_PRP_2_data.encoding = encoding;
}


/*****************************************************/

int GetFromStruct_SST_PRP_2_Encoding() {
  return SST_PRP_2_data.encoding;
}

/*****************************************************/

void OutputToFileASCII_EGG_NOM_2(FILE* fid) {
  if (fid) {
    fprintf(fid, "%20.9Lf   %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e   %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e   %d %d %d %d %d %d   %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e   %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e   %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e   %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e  %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e   %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e   %15.8e %15.8e %15.8e %15.8e\n", 
	    EGG_NOM_2_data.time,
	    EGG_NOM_2_data.Vxx,   EGG_NOM_2_data.Vyy,   EGG_NOM_2_data.Vzz,   EGG_NOM_2_data.Vxy,   EGG_NOM_2_data.Vxz,   EGG_NOM_2_data.Vyz,
	    EGG_NOM_2_data.Sxx,   EGG_NOM_2_data.Syy,   EGG_NOM_2_data.Szz,   EGG_NOM_2_data.Sxy,   EGG_NOM_2_data.Sxz,   EGG_NOM_2_data.Syz,
	    EGG_NOM_2_data.Fxx,   EGG_NOM_2_data.Fyy,   EGG_NOM_2_data.Fzz,   EGG_NOM_2_data.Fxy,   EGG_NOM_2_data.Fxz,   EGG_NOM_2_data.Fyz,
	    EGG_NOM_2_data.TTDxx, EGG_NOM_2_data.TTDyy, EGG_NOM_2_data.TTDzz, EGG_NOM_2_data.TTDxy, EGG_NOM_2_data.TTDxz, EGG_NOM_2_data.TTDyz, 
	    EGG_NOM_2_data.TTSxx, EGG_NOM_2_data.TTSyy, EGG_NOM_2_data.TTSzz, EGG_NOM_2_data.TTSxy, EGG_NOM_2_data.TTSxz, EGG_NOM_2_data.TTSyz, 
	    EGG_NOM_2_data.TTOxx, EGG_NOM_2_data.TTOyy, EGG_NOM_2_data.TTOzz, EGG_NOM_2_data.TTOxy, EGG_NOM_2_data.TTOxz, EGG_NOM_2_data.TTOyz, 
	    EGG_NOM_2_data.TTPxx, EGG_NOM_2_data.TTPyy, EGG_NOM_2_data.TTPzz, EGG_NOM_2_data.TTPxy, EGG_NOM_2_data.TTPxz, EGG_NOM_2_data.TTPyz, 
	    EGG_NOM_2_data.TNxx,  EGG_NOM_2_data.TNyy,  EGG_NOM_2_data.TNzz,  EGG_NOM_2_data.TNxy,  EGG_NOM_2_data.TNxz,  EGG_NOM_2_data.TNyz, 
	    EGG_NOM_2_data.Cxx,   EGG_NOM_2_data.Cyy,   EGG_NOM_2_data.Czz,   EGG_NOM_2_data.Cxy,   EGG_NOM_2_data.Cxz,   EGG_NOM_2_data.Cyz,
	    EGG_NOM_2_data.Q1,    EGG_NOM_2_data.Q2,    EGG_NOM_2_data.Q3,    EGG_NOM_2_data.Q4);
  } else {
    printf("TAIHEN: output file is not open!\n");
  }
  initialize_EGG_NOM_2_data(); // prepare for next line
}

void OutputToFileASCII_SST_PCV_2(FILE* fid) {
  if (fid) {
    fprintf(fid, "%6d %6d   %14.7e\n",
	    SST_PCV_2_data.row, SST_PCV_2_data.column, SST_PCV_2_data.Cov);
  } else {
    printf("TAIHEN: output file is not open!\n");
  }
  initialize_SST_PCV_2_data();
} 

void OutputToFileASCII_SST_PKI_2(FILE* fid) {
  double GPS_time;
  GPS_time = (greg2jul(SST_PKI_2_data.YY, SST_PKI_2_data.MM, SST_PKI_2_data.DoM, SST_PKI_2_data.hh, SST_PKI_2_data.mm, SST_PKI_2_data.ss)
	      - greg2jul(1980.0, 1.0, 6.0, 0.0, 0.0, 0.0)) * 24 * 3600;
  if (fid) {
    fprintf(fid, "%20.9f   %15.6f %15.6f %15.6f %15.6f\n",
    	    GPS_time, 
	    SST_PKI_2_data.X * SST_PKI_2_data.Position_Factor,
	    SST_PKI_2_data.Y * SST_PKI_2_data.Position_Factor,  
	    SST_PKI_2_data.Z * SST_PKI_2_data.Position_Factor, 
	    SST_PKI_2_data.Clk);
    //    fprintf(fid, "%4.0f %2.0f %2.0f %2.0f %2.0f %2.8f   %12.6f %12.6f %12.6f %12.6f\n",
    //	    SST_PKI_2_data.YY, SST_PKI_2_data.MM, SST_PKI_2_data.DoM, SST_PKI_2_data.hh, SST_PKI_2_data.mm, SST_PKI_2_data.ss,
    //	    SST_PKI_2_data.X,  SST_PKI_2_data.Y,  SST_PKI_2_data.Z, SST_PKI_2_data.Clk);
  } else {
    printf("TAIHEN: output file is not open!\n");
  }
  initialize_SST_PKI_2_data();
} 

void OutputToFileASCII_SST_PRD_2(FILE* fid) {
  double GPS_time;
  GPS_time = (greg2jul(SST_PRD_2_data.YY, SST_PRD_2_data.MM, SST_PRD_2_data.DoM, SST_PRD_2_data.hh, SST_PRD_2_data.mm, SST_PRD_2_data.ss)
	      - greg2jul(1980.0, 1.0, 6.0, 0.0, 0.0, 0.0)) * 24 * 3600;
  if (fid) {
    fprintf(fid, "%20.9f   %15.6f %15.6f %15.6f %15.6f   %12.6f %12.6f %12.6f %12.6f\n",
	    GPS_time,
    	    SST_PRD_2_data.X * SST_PRD_2_data.Position_Factor,
	    SST_PRD_2_data.Y * SST_PRD_2_data.Position_Factor,
	    SST_PRD_2_data.Z * SST_PRD_2_data.Position_Factor,
	    SST_PRD_2_data.Clk,
    	    SST_PRD_2_data.Xd * SST_PRD_2_data.Velocity_Factor, 
	    SST_PRD_2_data.Yd * SST_PRD_2_data.Velocity_Factor,
	    SST_PRD_2_data.Zd * SST_PRD_2_data.Velocity_Factor,
	    SST_PRD_2_data.Clkd);
      // fprintf(fid, "%4.0f %2.0f %2.0f %2.0f %2.0f %2.8f   %12.6f %12.6f %12.6f %12.6f   %12.6f %12.6f %12.6f %12.6f\n",
      //    SST_PRD_2_data.YY, SST_PRD_2_data.MM, SST_PRD_2_data.DoM, SST_PRD_2_data.hh, SST_PRD_2_data.mm, SST_PRD_2_data.ss,
      //    SST_PRD_2_data.X,  SST_PRD_2_data.Y,  SST_PRD_2_data.Z,  SST_PRD_2_data.Clk,
      //    SST_PRD_2_data.Xd, SST_PRD_2_data.Yd, SST_PRD_2_data.Zd, SST_PRD_2_data.Clkd);
   } else {
    printf("TAIHEN: output file is not open!\n");
  }
  initialize_SST_PRD_2_data();
} 

void OutputToFileASCII_SST_PRM_2(FILE* fid) {
  if (fid) {
    fprintf(fid, "%20.9f   %2.16f %2.15f %2.16f %2.16f\n", 
	    SST_PRM_2_data.reftime, SST_PRM_2_data.Q1, SST_PRM_2_data.Q2, SST_PRM_2_data.Q3, SST_PRM_2_data.Q4);
  } else {
    printf("TAIHEN: output file is not open!\n");
  }
  initialize_SST_PRM_2_data();
} 
