#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <getopt.h>

#include <libxml/xmlreader.h>

#include "helpers.h"
#include "data_handling.h"
#include "tags.h"
#include "base64.h"

#define PRODUCT_NONE                  0b00000000
#define PRODUCT_EGG_NOM_2             0b00000001
#define PRODUCT_SST_PCV_2             0b00000010
#define PRODUCT_SST_PKI_2             0b00000100
#define PRODUCT_SST_PRD_2             0b00001000
#define PRODUCT_SST_PRM_2             0b00010000
#define PRODUCT_SST_PRP_2             0b00100000

#define NODE_NONE                     0
#define NODE_ELEMENT                  1
#define NODE_TEXT                     3
#define NODE_ATTRIBUTE                2
#define NODE_CDATA                    4
#define NODE_ENTITYREFERENCE          5
#define NODE_ENTITY                   6
#define NODE_PROCESSINGINSTRUCTION    7
#define NODE_COMMENT                  8
#define NODE_DOCUMENT                 9
#define NODE_DOCUMENTTYPE            10
#define NODE_DOCUMENTFRAGMENT        11
#define NODE_NOTATION                12
#define NODE_WHITESPACE              13
#define NODE_SIGNIFICANTWHITESPACE   14
#define NODE_ENDELEMENT              15
#define NODE_ENDENTITY               16
#define NODE_XMLDECLARATION          17

FILE *fid_egg_nom, *fid_sst_pcv, *fid_sst_pki, *fid_sst_prd, *fid_sst_prm, *fid_sst_prp;
unsigned int product = PRODUCT_NONE;

// commandline parameters
static struct option long_options[] = {
    {"if",  required_argument, 0, 'i'},
    {"hdr", required_argument, 0, 'h'},
    {"dir", required_argument, 0, 'd'},
    {0, 0, 0, 0}
  };


static void processNode(xmlTextReaderPtr reader, const char *out_filbasename) {

  const xmlChar *name, *value, *row, *column, *encoding, *attr;
  int level = -1, type = 0,  empty = 0;
  int dec_len;
  char out_filname[256];
  char str_egg_nom[10] = "EGG_NOM_2", end_egg_nom[6] = "ascii",
    str_sst_pcv[10] = "SST_PCV_2", end_sst_pcv[6] = "ascii",
    str_sst_pki[10] = "SST_PKI_2", end_sst_pki[6] = "ascii",
    str_sst_prd[10] = "SST_PRD_2", end_sst_prd[6] = "ascii",
    str_sst_prm[10] = "SST_PRM_2", end_sst_prm[6] = "ascii",
    str_sst_prp[10] = "SST_PRP_2", end_sst_prp[4] = "pdf";

  name = xmlTextReaderConstName(reader);
  if (name == NULL)
    name = BAD_CAST "--";

  level = xmlTextReaderDepth(reader);
  value = xmlTextReaderConstValue(reader);
  
  empty = xmlTextReaderIsEmptyElement(reader);
  type  = xmlTextReaderNodeType(reader);

  switch (type) {
    // open tag
  case NODE_ELEMENT:
    if (!empty) {    // only handle the element if it is not empty! (e.g. skip "<test />" but not "<test></test>")
      switch (level) {
      case 0: TAG_flags.level_0 = GetNodeTagfromName(name, level); break;
      case 1: TAG_flags.level_1 = GetNodeTagfromName(name, level); break;
      case 2: TAG_flags.level_2 = GetNodeTagfromName(name, level); break;
      case 3: TAG_flags.level_3 = GetNodeTagfromName(name, level); break;
      case 4: TAG_flags.level_4 = GetNodeTagfromName(name, level); break;
      case 5: TAG_flags.level_5 = GetNodeTagfromName(name, level); break;
      case 6: TAG_flags.level_6 = GetNodeTagfromName(name, level); break;
      case 7: TAG_flags.level_7 = GetNodeTagfromName(name, level); break;
      case 8: TAG_flags.level_8 = GetNodeTagfromName(name, level); break;
      case 9: TAG_flags.level_9 = GetNodeTagfromName(name, level); break;
      default:
	printf("MESSAGE: something went wrong -- level > 9 reached!\n");
      }
      
      // open files for output
      // EGG_NOM_2
      if (((TAG_flags.level_0 == TAG_0_Data_Block) && 
	   (TAG_flags.level_1 == TAG_1_EGG_NOM_2) &&
	   (TAG_flags.level_2 == TAG_2_NONE)) && 
	  (product != PRODUCT_EGG_NOM_2)) { 
	product = PRODUCT_EGG_NOM_2;
	snprintf(out_filname, 254, out_filbasename, str_egg_nom, end_egg_nom);
	fid_egg_nom = fopen(out_filname, "w");
	printf("     output to: %s\n", out_filname);
      } 

      // SST_PCV_2
      if (((TAG_flags.level_0 == TAG_0_Data_Block) &&
	   (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	   (TAG_flags.level_2 == TAG_2_SST_PCV_2) &&
	   (TAG_flags.level_3 == TAG_3_NONE)) && 
	  (product != PRODUCT_SST_PCV_2)) {
	product = PRODUCT_SST_PCV_2;
	snprintf(out_filname, 254, out_filbasename, str_sst_pcv, end_sst_pcv);
	fid_sst_pcv = fopen(out_filname, "w");
	printf("     output to: %s\n", out_filname);
      } 

      // SST_PKI_2
      if (((TAG_flags.level_0 == TAG_0_Data_Block) &&
	   (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	   (TAG_flags.level_2 == TAG_2_SST_PKI_2) &&
	   (TAG_flags.level_3 == TAG_3_NONE)) && 
	  (product != PRODUCT_SST_PKI_2)) {
	product = PRODUCT_SST_PKI_2;
	snprintf(out_filname, 255, out_filbasename, str_sst_pki, end_sst_pki);
	fid_sst_pki = fopen(out_filname, "w");
	printf("     output to: %s\n", out_filname);
      } 

      // SST_PRD_2
      if (((TAG_flags.level_0 == TAG_0_Data_Block) &&
	   (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	   (TAG_flags.level_2 == TAG_2_SST_PRD_2) &&
	   (TAG_flags.level_3 == TAG_3_NONE)) && 
	  (product != PRODUCT_SST_PRD_2)) { 
	product = PRODUCT_SST_PRD_2;
	snprintf(out_filname, 255, out_filbasename, str_sst_prd, end_sst_prd);
	fid_sst_prd = fopen(out_filname, "w");
	printf("     output to: %s\n", out_filname);
      }

      // SST_PRM_2
      if (((TAG_flags.level_0 == TAG_0_Data_Block) &&
	   (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	   (TAG_flags.level_2 == TAG_2_SST_PRM_2) &&
	   (TAG_flags.level_3 == TAG_3_NONE)) && 
	  (product != PRODUCT_SST_PRM_2)) { 
	product = PRODUCT_SST_PRM_2;
	snprintf(out_filname, 254, out_filbasename, str_sst_prm, end_sst_prm);
	fid_sst_prm = fopen(out_filname, "w");
	printf("     output to: %s\n", out_filname);
      } 
      
      // SST_PRP_2
      if (((TAG_flags.level_0 == TAG_0_Data_Block) &&
	   (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	   (TAG_flags.level_2 == TAG_2_SST_PRP_2) &&
	   (TAG_flags.level_3 == TAG_3_NONE)) && 
	  (product != PRODUCT_SST_PRP_2)) { 
	product = PRODUCT_SST_PRP_2;
	snprintf(out_filname, 254, out_filbasename, str_sst_prp, end_sst_prp);
	fid_sst_prp = fopen(out_filname, "w");
	printf("     output to: %s\n", out_filname);
      } 
    } // if(!xmlTextReaderIsEmptyElement(reader))

    // tag has important attributes
    switch (product) {
    case PRODUCT_SST_PCV_2:    // row/column data is present as attributes of the tag <Covariance_Record>!
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PCV_2) &&
	  (TAG_flags.level_3 == TAG_3_List_of_Covariance_Records) &&
	  (TAG_flags.level_4 == TAG_4_Covariance_Record) &&
	  (TAG_flags.level_5 == TAG_5_NONE)) {
	row    = xmlTextReaderGetAttribute(reader, "row");
	column = xmlTextReaderGetAttribute(reader, "column");
	PutToStruct_SST_PCV_2(TAG_flags, TAG_extra_ROW_COLUMN, row, column, NULL);
      }
      break;

    case PRODUCT_SST_PKI_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PKI_2) &&
	  (TAG_flags.level_3 == TAG_3_List_of_SP3c_Records) &&
	  (TAG_flags.level_4 == TAG_4_SP3c_Record) &&
	  (TAG_flags.level_5 == TAG_5_List_of_Satellite_IDs) &&
	  (TAG_flags.level_6 == TAG_6_L15) &&
	  (TAG_flags.level_7 == TAG_7_Position) &&
	  (TAG_flags.level_8 == TAG_8_NONE)) {
	attr = xmlTextReaderGetAttribute(reader, "unit");
	if (attr != NULL)
	  if (strcmp(attr, "km") == 0) { // conversion factor km --> m
	    PutToStruct_SST_PKI_2_Position_Factor(1000.0);
	  }
      }
      break; 

    case PRODUCT_SST_PRD_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PRD_2) &&
	  (TAG_flags.level_3 == TAG_3_List_of_SP3c_Records) &&
	  (TAG_flags.level_4 == TAG_4_SP3c_Record) &&
	  (TAG_flags.level_5 == TAG_5_List_of_Satellite_IDs) &&
	  (TAG_flags.level_6 == TAG_6_L15)) {
	if ((TAG_flags.level_7 == TAG_7_Position) &&
	    (TAG_flags.level_8 == TAG_8_NONE)) {
	  attr = xmlTextReaderGetAttribute(reader, "unit");
	  if (attr != NULL)
	    if (strcmp(attr, "km") == 0) { // conversion factor km --> m
	      PutToStruct_SST_PRD_2_Position_Factor(1000.0);
	    }
	} else if ((TAG_flags.level_7 == TAG_7_Velocity) &&
		   (TAG_flags.level_8 == TAG_8_NONE)) {
	  attr = xmlTextReaderGetAttribute(reader, "unit");
	  if (attr != NULL)
	    if (strcmp(attr, "dm/s") == 0) { // conversion factor dm/s --> m/s
	      PutToStruct_SST_PRD_2_Velocity_Factor(0.1);
	    }
	}
      }
      break;

    case PRODUCT_SST_PRP_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PRP_2) &&
	  (TAG_flags.level_3 == TAG_3_List_of_PDF_Files) &&
	  (TAG_flags.level_4 == TAG_4_PDF_File) &&
	  (TAG_flags.level_5 == TAG_5_NONE)) { 
	attr = xmlTextReaderGetAttribute(reader, "encoding");
	PutToStruct_SST_PRP_2(TAG_flags, attr);
      }
      break;
    } // switch (product)
    break;
    
    // process value
  case NODE_TEXT: 
    value = xmlTextReaderConstValue(reader);
    
    switch (product) {
    case PRODUCT_EGG_NOM_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_EGG_NOM_2) &&
	  (TAG_flags.level_2 == TAG_2_List_of_GG_time_Records) && 
	  (TAG_flags.level_3 == TAG_3_GG_time_Record)) {
	PutToStruct_EGG_NOM_2(TAG_flags, value);
      } 
      break;

    case PRODUCT_SST_PCV_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PCV_2) &&
	  (TAG_flags.level_3 == TAG_3_List_of_Covariance_Records) &&
	  (TAG_flags.level_4 == TAG_4_Covariance_Record)) {
	PutToStruct_SST_PCV_2(TAG_flags, TAG_extra_VALUE, NULL, NULL, value);
      } 
      break;

    case PRODUCT_SST_PKI_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PKI_2) &&
	  (TAG_flags.level_3 == TAG_3_List_of_SP3c_Records) &&
	  (TAG_flags.level_4 == TAG_4_SP3c_Record)) {
	PutToStruct_SST_PKI_2(TAG_flags, value);
      } 
      break;

    case PRODUCT_SST_PRD_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PRD_2) &&
	  (TAG_flags.level_3 == TAG_3_List_of_SP3c_Records) &&
	  (TAG_flags.level_4 == TAG_4_SP3c_Record)) {
	PutToStruct_SST_PRD_2(TAG_flags, value);
      }
      break;

    case PRODUCT_SST_PRM_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PRM_2) &&
	  (TAG_flags.level_3 == TAG_3_List_of_Rotation_Records) &&
	  (TAG_flags.level_4 == TAG_4_Rotation_Record)) {
	PutToStruct_SST_PRM_2(TAG_flags, value);
      }
      break;

    case PRODUCT_SST_PRP_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PRP_2) &&
	  (TAG_flags.level_3 == TAG_3_List_of_PDF_Files) &&
	  (TAG_flags.level_4 == TAG_4_PDF_File)) {
	if (fid_sst_prp) {
	  if (GetFromStruct_SST_PRP_2_Encoding() == ENCODING_BASE64) {
	    encoding = base64_decode(value, strlen(value), &dec_len);
	    fwrite(encoding, dec_len, 1, fid_sst_prp);
	  } else {
	    printf("no encoding!\n");
	    fprintf(fid_sst_prp, "%s", value);
	  }
	  PutToStruct_SST_PRP_2_Encoding(ENCODING_NONE);
	}
      } 
      break;
    } // switch (product)
    break;

    // end tag
  case NODE_ENDELEMENT:
    switch (level) {
    case 0: TAG_flags.level_0 = TAG_0_NONE; break;
    case 1: TAG_flags.level_1 = TAG_1_NONE; break;
    case 2: TAG_flags.level_2 = TAG_2_NONE; break;
    case 3: TAG_flags.level_3 = TAG_3_NONE; break;
    case 4: TAG_flags.level_4 = TAG_4_NONE; break;
    case 5: TAG_flags.level_5 = TAG_5_NONE; break;
    case 6: TAG_flags.level_6 = TAG_6_NONE; break;
    case 7: TAG_flags.level_7 = TAG_7_NONE; break;
    case 8: TAG_flags.level_8 = TAG_8_NONE; break;
    case 9: TAG_flags.level_9 = TAG_9_NONE; break;
    default:
      printf("MESSAGE: something went wrong - level > 9 reached!\n");
    }

    switch (product) {
    case PRODUCT_EGG_NOM_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_EGG_NOM_2)) {
	if ((TAG_flags.level_2 == TAG_2_List_of_GG_time_Records) &&
	    (TAG_flags.level_3 == TAG_3_NONE)) { // we believe that higher levels are also TAG_x_NONE
	  OutputToFileASCII_EGG_NOM_2(fid_egg_nom);
	} else if ((TAG_flags.level_2 == TAG_2_NONE)) { // we believe that higher levels are also TAG_x_NONE
	  printf("          ... finished writing EGG_NOM_2.\n");
	  fclose(fid_egg_nom);
	  product = PRODUCT_NONE;
	}
      }
      break;

    case PRODUCT_SST_PCV_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PCV_2)) {
	if ((TAG_flags.level_3 == TAG_3_List_of_Covariance_Records) &&
	    (TAG_flags.level_4 == TAG_4_NONE)) {
	  OutputToFileASCII_SST_PCV_2(fid_sst_pcv);
	} else if ((TAG_flags.level_3 == TAG_3_NONE)) {
	  printf("          ... finished writing SST_PCV_2.\n");
	  fclose(fid_sst_pcv);
	  product = PRODUCT_NONE;
	}
      }
      break; 

    case PRODUCT_SST_PKI_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PKI_2)) {
	if ((TAG_flags.level_3 == TAG_3_List_of_SP3c_Records) &&
	    (TAG_flags.level_4 == TAG_4_NONE)) { // we believe that higher levels are also TAG_x_NONE
	  OutputToFileASCII_SST_PKI_2(fid_sst_pki); 
	} else if ((TAG_flags.level_3 == TAG_3_NONE)) { // we believe that higher levels are also TAG_x_NONE
	  printf("          ... finished writing SST_PKI_2.\n");
	  fclose(fid_sst_pki);
	  product = PRODUCT_NONE;
	}
      }
      break;

    case PRODUCT_SST_PRD_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PRD_2)) {
	if ((TAG_flags.level_3 == TAG_3_List_of_SP3c_Records) &&
	    (TAG_flags.level_4 == TAG_4_NONE)) { // we believe that higher levels are also TAG_x_NONE
	  OutputToFileASCII_SST_PRD_2(fid_sst_prd); 
	} else if ((TAG_flags.level_3 == TAG_3_NONE)) { // we believe that higher levels are also TAG_x_NONE
	  printf("          ... finished writing SST_PRD_2.\n");
	  fclose(fid_sst_prd);
	  product = PRODUCT_NONE;
	}
      }
      break;

    case PRODUCT_SST_PRM_2:
      if ((TAG_flags.level_0 == TAG_0_Data_Block) &&
	  (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	  (TAG_flags.level_2 == TAG_2_SST_PRM_2)) {
	if ((TAG_flags.level_3 == TAG_3_List_of_Rotation_Records) &&
	    (TAG_flags.level_4 == TAG_4_NONE)) {
	  OutputToFileASCII_SST_PRM_2(fid_sst_prm);
	} else if ((TAG_flags.level_3 == TAG_3_NONE)) {
	  printf("          ... finished writing SST_PRM_2.\n");
	  fclose(fid_sst_prm);
	  product = PRODUCT_NONE;
	}
      }
      break;
    case PRODUCT_SST_PRP_2:
      if (((TAG_flags.level_0 == TAG_0_Data_Block) &&
	   (TAG_flags.level_1 == TAG_1_SST_PSO_2) &&
	   (TAG_flags.level_2 == TAG_2_SST_PRP_2) &&
	   (TAG_flags.level_3 == TAG_3_NONE)) &&
	  (product == PRODUCT_SST_PRP_2)) {
	printf("          ... finished writing SST_PRP_2.\n");
	fclose(fid_sst_prp);
	product = PRODUCT_NONE;
      }
      break;
    } // switch (product)
    break;
    //  default:
    //    printf("MESSAGE: found tag type %d at level %d\n", type, level);
  } // switch (type)
}


static void streamFile(const char *in_filname, const char *out_filbasename, const char *working_dir) 
{
  xmlTextReaderPtr reader;
  int ret;
  char in_fil[256], out_filbase[256];

  snprintf(in_fil, 255, "%s%s", working_dir, in_filname);
  snprintf(out_filbase, 255, "%s%s", working_dir, out_filbasename);

  reader = xmlReaderForFile(in_fil, NULL, 0);
  if (reader != NULL) {
    ret = xmlTextReaderRead(reader);
    while (ret == 1) {
      processNode(reader, out_filbase);
      ret = xmlTextReaderRead(reader);
    }
    xmlFreeTextReader(reader);
    if (ret != 0) {
      fprintf(stderr, "TAIHEN: %s : failed to parse\n", in_fil);
    }
  } else {
    fprintf(stderr, "TAIHEN: Unable to open %s\n", in_fil);
  }
}

/* HEADER FILE ****************************************************/

static void processHdrNode(xmlTextReaderPtr reader) {
  const xmlChar *name, *value;
  int type = 0, empty = 0;

  name = xmlTextReaderConstName(reader);
  if (name == NULL)
    name = BAD_CAST "--";

  value = xmlTextReaderConstValue(reader);
  
  empty = xmlTextReaderIsEmptyElement(reader);
  type  = xmlTextReaderNodeType(reader);

  switch (type) {
  case NODE_ELEMENT:
    if (!empty) { // only handle the element if it is not empty! (e.g. skip "<test />" but not "<test></test>")
      HDR_TAG_flags |= HDR_GetNodeTagfromFile(name);
    }
    break;

  case NODE_TEXT:
    value = xmlTextReaderConstValue(reader);

    if ((HDR_TAG_flags & (HDR_TAG_Earth_Explorer_Header | HDR_TAG_Variable_Header | HDR_TAG_SPH | 
			  HDR_TAG_SST_PSO_2 | HDR_TAG_SST_PRM_2 | HDR_TAG_Epoch_Information | HDR_TAG_Reference))
	== (HDR_TAG_Earth_Explorer_Header | HDR_TAG_Variable_Header | HDR_TAG_SPH | 
	    HDR_TAG_SST_PSO_2 | HDR_TAG_SST_PRM_2 | HDR_TAG_Epoch_Information | HDR_TAG_Reference)) { 
      HDR_PutToStruct_SST_PRM_2(HDR_TAG_flags, value);
    } 
    break;

  case NODE_ENDELEMENT:
    HDR_TAG_flags &= ~HDR_GetNodeTagfromFile(name); 
    break;

  case NODE_SIGNIFICANTWHITESPACE:
  case NODE_CDATA:
    break;

  default:
    printf("MESSAGE: found type %d\n", type);
  } // switch (type)
}

static void streamHdrFile(const char *hdr_filname, const char *working_dir)
{
  xmlTextReaderPtr reader;
  int ret;
  char hdr_fil[256];

  snprintf(hdr_fil, 255, "%s%s", working_dir, hdr_filname);

  reader = xmlReaderForFile(hdr_fil, NULL, 0);
  if (reader != NULL) {
    ret = xmlTextReaderRead(reader);
    while (ret == 1) {
      processHdrNode(reader);
      ret = xmlTextReaderRead(reader);
    }
    xmlFreeTextReader(reader);
    if (ret != 0) {
      fprintf(stderr, "TAIHEN: %s : failed to parse\n", hdr_fil);
    }
  } else {
    fprintf(stderr, "TAIHEN: Unable to open %s\n", hdr_fil);
  }
}


/* ********** MAIN ********** */
int main(int argc, char *argv[])
{
  int command;
  int option_index = 0;

  xmlDoc *doc = NULL;
  char in_filname[256], hdr_filname[256], working_dir[256], out_filbasename[256];
  char partA[128], partB[128];

  in_filname[0]  = 0;
  hdr_filname[0] = 0;

  // read command line
  while ((command = getopt_long(argc, argv, "i:d:", long_options, &option_index)) != -1) {
    switch (command) {
    case 'i':
      strncpy(in_filname, optarg, 255);
      break;
    case 'h':
      strncpy(hdr_filname, optarg, 255);
      break;
    case 'd':
      strncpy(working_dir, optarg, 255);
      break;
    }
  }

  if (strlen(in_filname) == 0) {
    fprintf(stderr, "TAIHEN: please specify input filename!\n");
    return -1;
  } 
  printf("in_filname            : %s\n", in_filname);
  if (strlen(hdr_filname) == 0) {
    fprintf(stderr, "MESSAGE: header filename not specified - setting SST_PRM.Epoch_Information to 0.0!\n");
  } else {
    printf("header_filname        : %s\n", hdr_filname);
  }

  // form output filename:
  strncpy(partA, in_filname, 8); partA[8] = '\0';
  strncpy(partB, in_filname + 17, strpos(in_filname, ".") + 1 - 17); partB[strpos(in_filname, ".") + 1 - 17] = '\0';
  snprintf(out_filbasename, 255, "%s%%s%s%%s", partA, partB);
  
  // maybe here: check working_dir for end-"/" ...

  printf("output base file name : %s\n", out_filbasename);
  printf("working directory     : %s\n", working_dir);

  // initialize the variables
  initialize_EGG_NOM_2_data();
  initialize_SST_PCV_2_data();
  initialize_SST_PKI_2_data();
  initialize_SST_PRD_2_data();
  initialize_SST_PRM_2_data();
  HDR_initialize_SST_PRM_2_data();

  // parse file
  if (strlen(hdr_filname) != 0) { // header file
    streamHdrFile(hdr_filname, working_dir);
  }
  streamFile(in_filname, out_filbasename, working_dir);

  // free the global variables that may have been allocated by the parser.
  xmlCleanupParser();
  return 0;
}
